from typing import Generic, TypeVar

from omegaconf import DictConfig, OmegaConf

from pcs.init import initialize_object_nones

T = TypeVar("T")


class Component(Generic[T]):
    @classmethod
    def init_with_conf(
        cls: "type[Component[T]]", conf: DictConfig, Cls: type[T]
    ) -> "Component[T]":
        runtime = initialize_object_nones(Cls)
        return cls(conf, runtime)

    def __init__(self, conf: DictConfig, runtime: T):
        internal_attr_to_value = {
            "conf": conf,
            "runtime": runtime,
            "sealed": False,
        }
        conf_set = set(conf.keys())
        runtime_set = set(runtime.__dict__)
        internal_set = set(internal_attr_to_value.keys())

        def assert_no_duplicate_keys(set1: set, set2: set, error_message):
            duplicate_keys = set1.intersection(set2)
            assert (
                len(duplicate_keys) == 0
            ), f"Error initializing component - {error_message} - {duplicate_keys=}"

        assert_no_duplicate_keys(
            conf_set,
            runtime_set,
            "conf and runtime set cannot have keys which are named the same",
        )
        assert_no_duplicate_keys(
            internal_set,
            conf_set,
            "conf set has keys which are equal to internal component keys",
        )
        assert_no_duplicate_keys(
            internal_set,
            runtime_set,
            "runtime set has keys which are equal to internal component keys",
        )
        for k, v in internal_attr_to_value.items():
            super().__setattr__(k, v)

    def seal(self):
        OmegaConf.set_readonly(super().__getattribute__("conf"), True)
        super().__setattr__("sealed", True)

    def is_sealed(self):
        return self.sealed

    def get_conf(self, resolve=False):
        conf = super().__getattribute__("conf")
        if resolve:
            return OmegaConf.to_container(conf, resolve=True)
        return conf

    def get_runtime(self):
        return super().__getattribute__("runtime")

    def __getattr__(self, name):
        conf: DictConfig = super().__getattribute__("conf")
        if name in conf.keys():
            if name in conf:
                return getattr(conf, name)
            return None
        runtime: T = super().__getattribute__("runtime")
        if hasattr(runtime, name):
            return getattr(runtime, name)

    def __setattr__(self, name: str, item):
        if name in ["conf", "runtime"]:
            return super().__setattr__(name, item)
        conf: DictConfig = super().__getattribute__("conf")
        if name in conf.keys():
            return setattr(conf, name, item)
        runtime: T = super().__getattribute__("runtime")
        if hasattr(runtime, name):
            return setattr(runtime, name, item)
        raise AttributeError(f"{name} not found in class Component")

    def __repr__(self):
        conf = super().__getattribute__("conf")
        runtime = super().__getattribute__("runtime")
        sealed = super().__getattribute__("sealed")
        return f"Component({conf=}, {runtime=}, {sealed=})"

    def get_non_null_members_as_dict(self):
        resolved_conf = OmegaConf.to_container(
            super().__getattribute__("conf"), resolve=True
        )
        assert isinstance(resolved_conf, dict)
        result = {k: v for k, v in resolved_conf.items() if v != "???"}
        runtime = super().__getattribute__("runtime")
        result |= {k: v for k, v in runtime.__dict__.items() if v is not None}
        return result



import inspect
from collections.abc import Mapping
from typing import Any, Callable, TypeAlias, TypeVar

from pcs.component import Component

System: TypeAlias = Callable[..., dict[str, Any] | None]

T = TypeVar("T")


class Pipeline:
    def __init__(
        self,
        component: Component[T],
        systems: list[System],
        do_null_checks: bool = True,
        do_seal_check: bool = True,
    ):
        assert len(systems) > 0, "Systems in an empty list"
        self.component = component
        self.systems = systems
        self.do_null_checks = do_null_checks
        self.current_system = systems[0]
        self.do_seal_check = do_seal_check

    def execute(self) -> None:
        if self.do_seal_check:
            assert self.component.is_sealed(), (
                "You need to call seal on the component object to make the config objects static, or otherwise set `do_seal_check` to false for this pipeline"
            )
        for system in self.systems:
            self.current_system = system
            self.check_for_nulls(system)
            result = system(
                *[
                    getattr(self.component, param_name)
                    for param_name in inspect.signature(system).parameters
                ]
            )
            self.set_component(result)

    def check_for_nulls(self, system: System) -> None:
        if not self.do_null_checks:
            return
        for parameter in inspect.signature(system).parameters:
            assert getattr(self.component, parameter) is not None, (
                f"System: `{system.__name__}` - Parameter `{parameter}` was None"
            )

    def set_component(self, result: None | Mapping[str, object]) -> None:
        if result is None:
            return
        for name, obj in result.items():
            assert hasattr(self.component, name), (
                f"Component does not have an attribute called: `{name}`"
            )
            setattr(self.component, name, obj)



import argparse
from typing import TypeVar

from omegaconf import DictConfig, OmegaConf

from pcs.component import Component

T = TypeVar("T")


def parse_arguments_cli(ClsConf: object, ClsRealtime: type[T]) -> Component[T]:
    conf = OmegaConf.structured(ClsConf)
    args = do_parse_arguments()
    conf = update_dict_with_comma_separated_file_list(conf, args.args_files)
    assert isinstance(conf, DictConfig)
    conf = update_dict_with_rest_arguments(conf, args.rest)
    assert isinstance(conf, DictConfig)
    return Component.init_with_conf(conf, ClsRealtime)


def parse_arguments_from_files(
    ClsConf: object, ClsRealtime: type[T], files: list[str]
) -> Component[T]:
    conf = OmegaConf.structured(ClsConf)
    conf = update_dict_with_files(conf, files)
    assert isinstance(conf, DictConfig)
    return Component.init_with_conf(conf, ClsRealtime)


class Args(argparse.Namespace):
    def __init__(self):
        super().__init__()
        self.args_files: str = ""
        self.rest: list[str] = []


def do_parse_arguments() -> Args:
    parser = argparse.ArgumentParser()
    _ = parser.add_argument(
        "--args-files",
        "-f",
        required=False,
        help=(
            "A list of comma separated yaml files, where the keys in the "
            "latest files will overwrite those in previous files"
        ),
    )
    _ = parser.add_argument(
        "--rest",
        "-r",
        action="append",
        help=(
            "Set a number of key-value pairs like <key>=<value>. "
            "Values may have spaces if the whole value is wrapped"
            'in quotes such as <key>="this is a value". These arguments'
            "take precedence over those that come from files"
        ),
    )
    args = parser.parse_args(namespace=Args())
    return args


def update_dict_with_comma_separated_file_list(conf: DictConfig, files: str):
    return update_dict_with_files(
        conf, files.split(",") if len(files) > 0 else []
    )


def update_dict_with_files(conf: DictConfig, files: list[str]):
    return OmegaConf.merge(conf, *[OmegaConf.load(f) for f in files])


def update_dict_with_rest_arguments(conf: DictConfig, rest: list[str]):
    return OmegaConf.merge(conf, OmegaConf.from_dotlist(rest))
